/**
 * Created on Sep 11, 2003
 * @author Richard Gradischnegg RG
 */
package isac.bridge;

import isac.interfaces.ICalcIterator;
import isac.interfaces.IToCalc;
import isac.interfaces.IToUser;
import isac.util.CalcChanged;
import isac.util.AutoCalculateException;
import isac.util.formulae.CalcHead;
import isac.util.formulae.CalcFormula;
import isac.util.formulae.ContextMethod;
import isac.util.formulae.ContextProblem;
import isac.util.formulae.ContextTheory;
import isac.util.formulae.MethodID;
import isac.util.formulae.Position;
import isac.util.formulae.ProblemID;
import isac.util.tactics.Tactic;
import isac.wsdialog.WorksheetDialog;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * A CalcTree represents a Calculation: It is a tree which consists of one
 * CalcHead at the top and abitrarily many Formulas and SubCalculations (with
 * separate CalcHeads). This is the main interface between the Bridge and the
 * Dialog. The CalcTree calls methods of the Math Engine, which in turn are
 * delegated to the BridgeRMI object on the other side of the RMI connection
 * ??????WN0409. The communication with the SML Kernel is completly hided from
 * the Dialog and the Frontend.
 * 
 * @author Richard Gradischnegg
 */

public class CalcTree implements IToCalc, Serializable {

    /**
     * For serialization, see corresponding API documentation
     */
    private static final long serialVersionUID = 501L;

    private static final Logger logger = Logger.getLogger(CalcTree.class
            .getName());

    private int id_;

    private Vector listeners_;

    private MathEngine math_engine_;

    private ICalcIterator hot_spot_;// WN040924 ..delete: control hotSpot --> DG

    private IBridgeRMI bridge_rmi_;// WN040924 added for completeCalcHead & Co.

    /**
     * Start a new calculation. In case of failure, the constructor may change
     * the formalization object to hint at incomplete or contradictory
     * specifications. <br>
     * Only complete calcHeads should be passed. <br>
     * Check every CalcHead for completeness (with the static method
     * checkcalcHead() before constructing a CalcTree ..TODO WN040824 shift to
     * authoring-tools
     * 
     * @param identification
     *            for the CalcTree as delivered by the SML-kernel
     */
    CalcTree(MathEngine me, int id) {
        this.id_ = id;
        this.math_engine_ = me;
        this.bridge_rmi_ = me.getBridgeRMI();// WN040924
        listeners_ = new Vector();
    }

    /**
     * Attach another dialog to the existing calculation tree.
     * 
     * @param dlg
     *            The new Dialog to be attached to this CalcTree
     */
    public void attach(WorksheetDialog dg) {
        listeners_.add(dg);
    }

    /**
     * Get a new iterator referencing the root element of the calculation tree.
     * <p>
     * The functionality for navigating the calculation is implemented in the
     * CalcIterator object.
     * 
     * @return new instance of CalcIterator
     */
    public ICalcIterator iterator() {
        return new CalcIterator(this);
    }

    /**
     * Makes the passed formula the active formula, i.e. the formula the next
     * tactic is applied to.
     * 
     * @param newActiveFormula
     */
    public void moveActiveFormula(ICalcIterator cit) {
        try {
            // Position p =
            math_engine_.moveActiveFormula(id_, cit.getPosition());
            // if (p == null)//WN0412 message ?!
            // return false;
            // else
            // return true;
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    /**
     * start a calculation which throws a CalcChanged in the MathEngine
     * 
     * @return transactionID WN0507 unused
     */
    public int startCalculate() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: startCalculate()");
        CChanged cc;
        int transactionID = -1;
        try {
            cc = math_engine_.startCalculate(this.id_);
            if (!(cc == null)) {
                hot_spot_ = new CalcIterator(this, cc.getLastGenerated());
                transactionID = generateTransactionID();
//                informListeners(new CalcIterator(this, cc.getLastUnchanged()),
//                        new CalcIterator(this, cc.getLastDeleted()),
//                        new CalcIterator(this, cc.getLastGenerated()),
//                        transactionID);
                CalcChanged calcc = new CalcChanged(transactionID,
                        true, // completed
                        new CalcIterator(this, cc.getLastUnchanged()),
                        new CalcIterator(this, cc.getLastDeleted()),
                        new CalcIterator(this, cc.getLastGenerated()));
                informListeners(calcc);
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return transactionID;
    }

    /**
     * @see isac.interfaces.IToCalc#modelProblem() WN050809 push this method
     *      through to the SML-kernel
     */
    public CalcHead modelProblem() throws RemoteException {
        return math_engine_.modelProblem(this.id_);
    }

    /**
     * @see isac.interfaces.IToCalc#nextSpecify() WN050809 push this method
     *      through to the SML-kernel
     */
    public CalcHead nextSpecify() throws RemoteException {
        return math_engine_.nextSpecify(this.id_);
    }

    /**
     * @see isac.interfaces.IToCalc#checkCalcHead(isac.util.formulae.CalcHead)
     */
    public CalcHead checkCalcHead(CalcHead calcHead) throws RemoteException {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: modifyCalcHead(" + calcHead.toSMLString());
        return (CalcHead) math_engine_.checkCalcHead(this.id_, calcHead);
        // WN040924 calc_head_.fillValuesfrom(newCalcHead);
    }

    /**
     * @see isac.interfaces.IToCalc#completeModel() WN050809 push this method
     *      through to the SML-kernel
     */
    public CalcHead completeModel() throws RemoteException {
        return math_engine_.completeModel(this.id_);
    }

    /**
     * @see isac.interfaces.IToCalc#setTheory()
     */
    public CalcHead setTheory(String thy_id) throws RemoteException {
        return math_engine_.setTheory(this.id_, thy_id);
    }

    /**
     * @see isac.interfaces.IToCalc#setProblem()
     */
    public CalcHead setProblem(ProblemID id) throws RemoteException {
        return math_engine_.setProblem(this.id_, id);
    }

    /**
     * @see isac.interfaces.IToCalc#setMethod()
     */
    public CalcHead setMethod(MethodID id) throws RemoteException {
        return math_engine_.setMethod(this.id_, id);
    }

    /**
     * @see isac.interfaces.IToCalc#completeCalcHead()
     */
    public CalcHead completeCalcHead() throws RemoteException {
        if (logger.isInfoEnabled())
            logger.info("  DG<>BR: completeCalcHead()");
        return bridge_rmi_.completeCalcHead(this.id_);
    }

    /**
     * @see isac.interfaces.IToCalc#completeCalcHead(isac.util.formulae.CalcHead,
     *      int)
     * @deprecated because not yet implemented in SML
     */
    public CalcHead completeCalcHead(CalcHead calcHead, int completeItem) {
        // WN0508 left as a model for modeling CalcHead as an object-reference
        // CalcHead newCalcHead;
        // try {
        // //WN040924 newCalcHead = bridgeRMI.completeCalcHead(calc_head_,
        // // completeItem);
        // bridge_rmi_.completeCalcHead(this.id_, calc_head_, completeItem);
        // //WN040924 calc_head_.fillValuesfrom(newCalcHead);
        // calc_head_.setCalcHeadStatus(CalcHead.MODEL_ITEM_CORRECT);
        //
        // } catch (RemoteException e) {
        // e.printStackTrace();
        // }
        return null;
    }

    /**
     * @see isac.interfaces.IToCalc#resetCalcHead()
     */
    public CalcHead resetCalcHead() throws RemoteException {
        return math_engine_.resetCalcHead(this.id_);
    }

    /**
     * @see isac.interfaces.IToCalc#startSolving()
     */
    public void startSolving() throws AutoCalculateException {
        Tactic tac = this.fetchProposedTactic();
        if (tac.getName().compareTo("Apply_Method") == 0)
            this.autoCalculate(IToCalc.SCOPE_CALCULATION, 1);
        else
            new AutoCalculateException(tac);
    }

    /**
     * Calculate the next n steps starting from the active formula. The Tactic
     * to be applied in this step may be preset by setNextTactic
     * 
     * @param scope
     *            The scope parameter can have following values:
     *            <ul>
     *            <li>Stay in the current subproblem
     *            <li>Stay in the current subcalculation (e.g. repetitions)
     *            <li>Stay in the current calculation as a whole.
     *            </ul>
     *            Constants for this scopes are found in the IToCalc class
     * @param nSteps
     *            Specifies the count of steps to be autocalculated. Passing 0
     *            for nSteps will calculate until reaching a final result.
     * @return The method returns an unique id to identify the request when
     *         notifying about updates in the tree.
     * 
     * WN0504 see differences to this method completely done by RG !
     */
    public int autoCalculate(int scope, int nSteps) {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: autoCalculate(" + scope + ", " + nSteps
                    + ")");
        CChanged cc;
        int transactionID = -1;
        try {
            cc = math_engine_.autoCalculate(this.id_, scope, nSteps);
            if (!(cc == null)) {
                hot_spot_ = new CalcIterator(this, cc.getLastGenerated());
                transactionID = generateTransactionID();
//                informListeners(new CalcIterator(this, cc.getLastUnchanged()),
//                        new CalcIterator(this, cc.getLastDeleted()),
//                        new CalcIterator(this, cc.getLastGenerated()),
//                        transactionID);
                CalcChanged calcc = new CalcChanged(transactionID,
                        true, // completed
                        new CalcIterator(this, cc.getLastUnchanged()),
                        new CalcIterator(this, cc.getLastDeleted()),
                        new CalcIterator(this, cc.getLastGenerated()));
                informListeners(calcc);
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return transactionID;
    }

    /**
     * Tries to replaces the active formula in the calculation tree by the
     * passed formula. This could fail if no valid derivation from the preceding
     * formula to the (new) active formula can be found.
     * <p>
     * Parts of the calculation after the replaced formula are likely to become
     * invalid, listeners will be informed of this with the update message,
     * which contains the last unchanged formula.
     * 
     * @param newFormula
     * @return The method will return 0 on success or a nonzero value indicating
     *         the status otherwise.
     */
    public int replaceFormula(CalcFormula f) throws RemoteException {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: replaceFormula(" + f.toSMLString() + ")");
        CChanged cc;
        int transactionID = -1;
        try {
            cc = math_engine_.replaceFormula(this.id_, f);
            if (!(cc == null)) {
                hot_spot_ = new CalcIterator(this, cc.getLastGenerated());
                transactionID = generateTransactionID();
//                informListeners(new CalcIterator(this, cc.getLastUnchanged()),
//                        new CalcIterator(this, cc.getLastDeleted()),
//                        new CalcIterator(this, cc.getLastGenerated()),
//                        transactionID);
                CalcChanged calcc = new CalcChanged(transactionID,
                        true, // completed
                        new CalcIterator(this, cc.getLastUnchanged()),
                        new CalcIterator(this, cc.getLastDeleted()),
                        new CalcIterator(this, cc.getLastGenerated()));
                informListeners(calcc);
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return 0;
    }

    /**
     * Tries to append the passed formula after the active formula. This could
     * fail if no valid derivation from the preceding formula to the (new)
     * active formula can be found. The method will fail as well if the active
     * formula is not the last formula in the (sub)calculation. Parts of the
     * calculation might become invalid.
     * 
     * @param newFormula
     * @return The method will return 0 on success or a nonzero value indicating
     *         the status otherwise.
     */
    public int appendFormula(CalcFormula f) throws RemoteException {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: appendFormula(" + f.toSMLString() + ")");
        CEvent ce;
        int transactionID = -1;
        try {
            ce = math_engine_.appendFormula(this.id_, f);
            if (!(ce == null)) {
                if (ce instanceof CChanged) {
                    CChanged cc = (CChanged) ce;
                    hot_spot_ = new CalcIterator(this, cc.getLastGenerated());
                    transactionID = generateTransactionID();
                    // informListeners(new CalcIterator(this, cc
                    // .getLastUnchanged()), new CalcIterator(this, cc
                    // .getLastDeleted()), new CalcIterator(this, cc
                    // .getLastGenerated()), transactionID);
                    CalcChanged calcc = new CalcChanged(transactionID,
                            true, // completed
                            new CalcIterator(this, cc.getLastUnchanged()),
                            new CalcIterator(this, cc.getLastDeleted()),
                            new CalcIterator(this, cc.getLastGenerated()));
                    informListeners(calcc);
                } else {
                    CMessage cm = (CMessage) ce;
                    // informListeners(new )
                }
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return 0;
    }

    /**
     * @see isac.interfaces.IToCalc#intermediateSteps()
     */
    public int intermediateSteps(ICalcIterator ci) throws RemoteException {
        if (logger.isInfoEnabled())
            logger
                    .info("  DG->BR: intermediateSteps(" + ci.toSMLString()
                            + ")");
        CChanged cc;
        int transactionID = -1;
        try {
            cc = math_engine_.intermediateSteps(this.id_, ci);
            if (!(cc == null)) {
                transactionID = generateTransactionID();
//                informListeners(new CalcIterator(this, cc.getLastUnchanged()),
//                        new CalcIterator(this, cc.getLastDeleted()),
//                        new CalcIterator(this, cc.getLastGenerated()),
//                        transactionID);
                CalcChanged calcc = new CalcChanged(transactionID,
                        true, // completed
                        new CalcIterator(this, cc.getLastUnchanged()),
                        new CalcIterator(this, cc.getLastDeleted()),
                        new CalcIterator(this, cc.getLastGenerated()));
                informListeners(calcc);
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return 0;
    }

    /**
     * Set a tactic for usage by autocalculate.
     * 
     * @param tactic
     *            the Tactic to be used in the the next step
     * @return The return value indicates success, failure or probable success
     *         at a later time.
     */
    public int setNextTactic(Tactic tactic) {
        if (logger.isInfoEnabled())
            logger
                    .info("  DG->BR: setNextTactic(" + tactic.toSMLString()
                            + ")");
        // int result = -1;
        try {
            // result =
            math_engine_.setNextTactic(this.id_, tactic);

            /*******************************************************************
             * Old code ... We do not know what this does.
             * 
             */
            // if (result == 0) {
            // hot_spot_.moveDown(); // move to the new formula/calc_head_
            // informListeners((ICalcIterator) hot_spot_.clone(),
            // // FIXXXME.WN0411: nothing changed, but if outcommented: subpbl
            // // shows
            // // Add_Given etc. ... discuss with LK !
            // (ICalcIterator) hot_spot_.clone(),
            // (ICalcIterator) hot_spot_.clone(),
            // generateTransactionID());
            // }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return -1;
    }

    /**
     * Retrieve the tactic proposed by the SML-Kernel for the next step in the
     * calculation.
     * 
     * @return proposed tactic. Can be null, if the kernel does not know what to
     *         do next, or if the end of the calculation is reached
     */
    public Tactic fetchProposedTactic() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: fetchProposedTactic()");
        try {
            Tactic tac = math_engine_.fetchProposedTactic(this.id_);
            if (logger.isInfoEnabled())
                logger.info("  DG<-BR: fetchProposedTactic <- "
                        + tac.toSMLString());
            return tac;
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: fetchProposedTactic <- NULL helpless");
        return null;
    }

    /**
     * Add a Listener to this CalcTree. Listeners will be informed about any
     * changes in the CalcTree.
     * 
     * @param listener
     *            the new Listener, has to implement the interface IToUser
     * @return true, if the listener has been added successfully
     */
    public boolean addDataChangeListener(IToUser listener) {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: addDataChangeListener(" + listener + ")");
        if (listeners_.contains(listener)) {
            return false;
        } else {
            listeners_.add(listener);
        }
        return true;
    }

    /**
     * Retrieve the active formula in a CalcTree. This is the formula form where
     * the calculation will continue.
     * 
     * @return Iterator that marks the currently active formula
     */
    public ICalcIterator getActiveFormula() {
        try {
            if (logger.isInfoEnabled())
                logger.info("  DG<>BR: getActiveFormula <- "
                        + hot_spot_.toSMLString());
            return (ICalcIterator) hot_spot_.cloneIterator();
        } catch (RemoteException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            return null;
        }

    }

    /**
     * Destruct this CalcTree: the sml-Representation of the CalcTree is
     * disposed, the id can be used for a new CalcTree
     * 
     * @return boolean value: success of operation (complete/failed)
     */
    public boolean destruct() {
        try {
            return math_engine_.destruct(this.id_);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return false;
    }

    // send an update event to the registered listeners
//  private void informListeners(ICalcIterator lastUnchangedFormula,
//  ICalcIterator lastDeletedFormula,
//  ICalcIterator lastGeneratedFormula, int transactionID)
  private void informListeners(CalcChanged calcc)
            throws RemoteException {
        if (logger.isInfoEnabled())
            logger.info(" . . <-BR: informListeners (unc="
                    + calcc.getLastUnchangedFormula().toSMLString() + ", del="
                    + calcc.getLastDeletedFormula().toSMLString() + ", gen="
                    + calcc.getLastGeneratedFormula().toSMLString());
        Iterator listn_it = listeners_.iterator();
        // WorksheetDialog dg;
        IToUser dg;
//        CalcChanged e = new CalcChanged(transactionID, true, // completed
//                lastUnchangedFormula, lastDeletedFormula, lastGeneratedFormula);
        while (listn_it.hasNext()) {
            dg = (IToUser) listn_it.next();
            dg.calcResponse(calcc);
        }
    }

    MathEngine getMathEngine() {
        return math_engine_;
    }

    public int getId() {
        return id_;
    }

    // Special iterator hotSpot: marks the currently active formula of
    // the calcTree
    void setHotSpot(ICalcIterator iterator) {
        hot_spot_ = iterator;
    }

    // WN0501 ... _NO_: ... = return-value of autoCalculate,etc. =
    // = CalcChanged.transaction_id_
    private int generateTransactionID() {
        return (int) (Math.random() * 0xFFFF);
    }

    public CalcHead setContext(int javaCalcID, ContextProblem context,
            Position pos) throws RemoteException {
        return math_engine_.setContext(javaCalcID, context, pos);
    }

    public CalcHead setContext(int javaCalcID, ContextMethod context,
            Position pos) throws RemoteException {
        return math_engine_.setContext(javaCalcID, context, pos);
    }

    public void setContext(int javaCalcID, ContextTheory context, Position pos)
            throws RemoteException {
        CChanged cc = math_engine_.setContext(javaCalcID, context,
                pos);

        int transactionID = -1;
        try {
            if (!(cc == null)) {
                hot_spot_ = new CalcIterator(this, cc
                        .getLastGenerated());
                transactionID = generateTransactionID();
//                informListeners(new CalcIterator(this, cald_changed
//                        .getLastUnchanged()), new CalcIterator(this,
//                        cald_changed.getLastDeleted()), new CalcIterator(this,
//                        cald_changed.getLastGenerated()), transactionID);
                CalcChanged calcc = new CalcChanged(transactionID,
                        true, // completed
                        new CalcIterator(this, cc.getLastUnchanged()),
                        new CalcIterator(this, cc.getLastDeleted()),
                        new CalcIterator(this, cc.getLastGenerated()));
                informListeners(calcc);
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

}