/*
 * @author Georg Kompacher, member of the ISAC-team, 
 * Copyright (c) 2006 by Georg Kompacher
 * created Jul 10, 2006 05:44:52 PM
 * Institute for Softwaretechnology, Graz University of Technology, Austria.
 * 
 * Use is subject to PGPL license terms.
 */
package isac.browserdialog;

import isac.gui.util.HierarchyNode;
import isac.interfaces.IToGUI;
import isac.session.Session;
import isac.useractions.EUIContext;
import isac.useractions.EUIElement;
import isac.useractions.IUserAction;
import isac.useractions.UIAction;
import isac.useractions.UserActionOnLink;
import isac.util.formulae.Context;
import isac.util.formulae.ContextUnknown;
import isac.util.formulae.HierarchyKey;
import isac.wsdialog.DialogProtocolException;
import isac.wsdialog.IContextProvider;
import isac.wsdialog.WorksheetState;
import isac.wsdialog.IContextProvider.ContextType;

import java.net.URL;
import java.rmi.RemoteException;
import java.util.Random;

import javax.swing.tree.TreePath;

/**
 * @author Georg Kompacher
 * 
 * This class is responsible for the interactions to the TheoryBrowser, to the
 * active WSDialog, and also to other BrowserDialogs.
 */
public class TheoryDialog extends BrowserDialog {

    /**
     * For serialization, see corresponding API Documentation; value generated
     * by eclipse to avoid a warning
     */
    private static final long serialVersionUID = 6479711601158544388L;

    /** true if the <To Worksheet> button is visible * */
    private boolean to_worksheet_visible_;

    /** true if the <Context On> button is visible* */
    // private boolean context_on_visible_;
    /**
     * true if the calc head of the active worksheet is open this flag is
     * updated in updateState()
     */
    private boolean calc_head_open_;

    /**
     * true if a worksheet is open this flag is updated in updateState()
     */
    private boolean worksheet_open_;

    /**
     * The constructor of this class which is given the session the TheoryDialog
     * is instanciated for. We need the session to get the correct
     * WSDialogManager (and also the active WSDialog) for this BrowserDialog.
     * 
     * @param session
     * @throws RemoteException
     */
    public TheoryDialog(Session session) throws RemoteException {
        super(session);
        context_type_ = ContextType.CONTEXT_THEORIES;

        hierarchy_ = loadHierarchy(context_type_);
        current_context_ = new ContextUnknown();
    }

    /**
     * Notify the TheoryDialog about events or requests triggered by the user.
     * The TheoryBrowser calls this method over RMI.
     * 
     * @param action
     *            What's new from the UI
     * @return true, if the event is being processed, false if processing is
     *         denied or impossible
     * @throws RemoteException
     * @throws DialogProtocolException
     */
    public boolean notifyUserAction(IUserAction action)
            throws DialogProtocolException, RemoteException {
        System.out.println("TheoryDialog::notifyUserAction() action="
                + action.getActionID());
        if (super.notifyUserAction(action))
            return true;

        user_logger_.logUserAction("thy", action);

        switch (action.getActionID()) {
        case UI_BROWSER_MATCH_ON: {
            context_on_ = true;
            drawButtons();
            sendInitContextToBrowser();
            return true;
        }
        case UI_BROWSER_MATCH_OFF: {
            context_on_ = false;
            drawButtons();
            ContextUnknown new_context = new ContextUnknown();
            new_context.setHierarchyKey(current_context_.getHierarchyKey());
            new_context.setKEStoreKey(current_context_.getKEStoreKey());
            sendContextToBrowser(new_context);
            sendLinkToMiniBrowser(new_context, context_type_);
            return true;
        }
        case UI_BROWSER_TO_WORKSHEET: {
            setCurrentContextToWorksheet();
            return true;
        }
        case UI_MINIBROWSER_LINK: {
            URL receivedLink = ((UserActionOnLink) action).getLink();
            try {
                interpretLink(receivedLink);
                return true;
            } catch (RemoteException e) {
                e.printStackTrace();
                return false;
            }
        }
        }
        return false;
    }

    /**
     * removes the button for To Worksheet. If the button does not exist so far,
     * nothing happens.
     */
    private void removeButtonToWorksheet() {
        if (!to_worksheet_visible_)
            return;
        try {
            browser_frame_rmi_.removeUIElement(new UIAction(
                    EUIElement.UI_BROWSER_TO_WORKSHEET,
                    EUIContext.UI_CONTEXT_BROWSER, user_language_));
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        to_worksheet_visible_ = false;
    }

    /**
     * generates the button for To Worksheet. If the button does already exist,
     * nothing happens.
     */
    private void drawButtonToWorksheet() {
        if (to_worksheet_visible_)
            return;
        try {
            browser_frame_rmi_.addUIElement(new UIAction(
                    EUIElement.UI_BROWSER_TO_WORKSHEET,
                    EUIContext.UI_CONTEXT_BROWSER, user_language_));
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        to_worksheet_visible_ = true;
    }

    /**
     * This method overrides the method in the base class and calls the method
     * of the base class. This is done, as some additional initialisations have
     * to be done with the buttons and the context.
     * 
     * @param browser_frame
     *            the browser frame to register
     */
    public void registerBrowserFrame(IToGUI browser_frame)
            throws RemoteException {
        super.registerBrowserFrame(browser_frame);
        initButtons();
    }

    private void initButtons() {
        context_on_visible_ = false;
        to_worksheet_visible_ = false;
        context_on_ = true;
        try {
            browser_frame_rmi_.addUIElement(new UIAction(
                    EUIElement.UI_BROWSER_MATCH_OFF,
                    EUIContext.UI_CONTEXT_BROWSER, user_language_));
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    /**
     * This method draws the buttons of the browser. Which buttons need to be
     * shown depends on three flags: context_on_ worksheet_open_ calc_head_open_ |
     * Buttons to show:
     * ----------------------------------------------------------------- F X X |
     * <Context Off> T F F | <Context On> T F T | impossible T T F | <Context
     * On> <ToWorksheet> T T T | <Context On>
     * 
     */

    protected void drawButtons() {
        if (!context_on_) {
            drawButtonContextOn();
            removeButtonToWorksheet();
        } else {
            if (worksheet_open_ && !calc_head_open_) {
                drawButtonContextOff();
                drawButtonToWorksheet();
            } else {
                drawButtonContextOff();
                removeButtonToWorksheet();
            }
        }
    }

    /**
     * This method does the update for the two variables: worksheet_open_ and
     * calc_head_open_. An interface to the active worksheet is fetched from the
     * session. If no active worksheet exists (and null is returned) there is no
     * open worksheet. The interface returns a WorksheetState which is able to
     * tell us, if the CalcHead is open.
     */
    public void updateState() {

        try {
            IContextProvider context_provider = session_.getContextProvider();
            if (context_provider == null) {
                worksheet_open_ = false;
                calc_head_open_ = false;
                return;
            }
            worksheet_open_ = true;
            WorksheetState state = context_provider.getWorksheetState();
            calc_head_open_ = state.isCalcHeadOpen();
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

}